// Copyright 2019, The Go Authors. All rights reserved.
// Use of this source code is governed by a BSD-style
// license that can be found in the LICENSE.md file.

package flags

// Deterministic controls whether the output of Diff should be deterministic.
// This is only used for testing.
var Deterministic bool
